# -*- coding: utf-8 -*-

# Copyright 2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of mongomotor.

# mongomotor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mongomotor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mongomotor. If not, see <http://www.gnu.org/licenses/>.

import gridfs
from gridfs import grid_file
from motor.metaprogramming import create_class_with_framework
from motor.motor_gridfs import (AgnosticGridFSBucket, AgnosticGridIn,
                                AgnosticGridOut, AgnosticGridOutCursor)
from mongomotor.core import (MongoMotorAgnosticDatabase,
                             MongoMotorAgnosticCollection)
from mongomotor.metaprogramming import OriginalDelegate


class MongoMotorAgnosticGridOut(AgnosticGridOut):
    """Class to read data out of GridFS.

    MotorGridOut supports the same attributes as PyMongo's
    :class:`~gridfs.grid_file.GridOut`, such as ``_id``, ``content_type``,
    etc.

    You don't need to instantiate this class directly - use the
    methods provided by :class:`~mongomotor.MotorGridFSBucket`. If it **is**
    instantiated directly, call :meth:`open`, :meth:`read`, or
    :meth:`readline` before accessing its attributes.
    """
    __motor_class_name__ = 'MongoMotorGridOut'

    def __init__(self, root_collection, file_id=None, file_document=None,
                 delegate=None):
        collection_class = create_class_with_framework(
            MongoMotorAgnosticCollection, self._framework, self.__module__)

        if not isinstance(root_collection, collection_class):
            raise TypeError(
                "First argument to MongoMotorGridOut must be "
                "MongoMotorCollection, not %r" % root_collection)

        if delegate:
            self.delegate = delegate
        else:
            self.delegate = self.__delegate_class__(
                root_collection.delegate,
                file_id,
                file_document)

        self.io_loop = root_collection.get_io_loop()


class MongoMotorAgnosticGridIn(AgnosticGridIn):

    __motor_class_name__ = 'MongoMotorGridIn'

    def __init__(self, root_collection, delegate=None, **kwargs):
        """
        Class to write data to GridFS. Application developers should not
        generally need to instantiate this class - see
        :meth:`~mongomotor.MongoMotorGridFSBucket.open_upload_stream`.

        Any of the file level options specified in the `GridFS Spec
        <http://dochub.mongodb.org/core/gridfs>`_ may be passed as
        keyword arguments. Any additional keyword arguments will be
        set as additional fields on the file document. Valid keyword
        arguments include:

          - ``"_id"``: unique ID for this file (default:
            :class:`~bson.objectid.ObjectId`) - this ``"_id"`` must
            not have already been used for another file

          - ``"filename"``: human name for the file

          - ``"contentType"`` or ``"content_type"``: valid mime-type
            for the file

          - ``"chunkSize"`` or ``"chunk_size"``: size of each of the
            chunks, in bytes (default: 256 kb)

          - ``"encoding"``: encoding used for this file. In Python 2,
            any :class:`unicode` that is written to the file will be
            converted to a :class:`str`. In Python 3, any :class:`str`
            that is written to the file will be converted to
            :class:`bytes`.

        :Parameters:
          - `root_collection`: A :class:`~mongomotor.MongoMotorCollection`,
             the root collection to write to.
          - `delegate`: An instance of the delegate class.
          - `**kwargs` (optional): file level options (see above)

        """
        collection_class = create_class_with_framework(
            MongoMotorAgnosticCollection, self._framework, self.__module__)

        if not isinstance(root_collection, collection_class):
            raise TypeError(
                "First argument to MotorGridIn must be "
                "MongoMotorCollection, not %r" % root_collection)

        self.io_loop = root_collection.get_io_loop()
        if delegate:
            # Short cut.
            self.delegate = delegate
        else:
            self.delegate = self.__delegate_class__(
                root_collection.delegate,
                **kwargs)


class MongoMotorAgnosticGridFS(AgnosticGridFSBucket):
    """Create a handle to a GridFS bucket.

    Raises :exc:`~pymongo.errors.ConfigurationError` if `write_concern`
    is not acknowledged.

    This class conforms to the `GridFS API Spec
    <https://github.com/mongodb/specifications/blob/master/source/gridfs/gridfs-spec.rst>`_
    for MongoDB drivers.

    :Parameters:
      - `database`: database to use.
      - `bucket_name` (optional): The name of the bucket. Defaults to 'fs'.
      - `chunk_size_bytes` (optional): The chunk size in bytes. Defaults
        to 255KB.
      - `write_concern` (optional): The
        :class:`~pymongo.write_concern.WriteConcern` to use. If ``None``
        (the default) db.write_concern is used.
      - `read_preference` (optional): The read preference to use. If
        ``None`` (the default) db.read_preference is used.
      - `disable_md5` (optional): When True, MD5 checksums will not be
        computed for uploaded files. Useful in environments where MD5
        cannot be used for regulatory or other reasons. Defaults to False.

    .. mongodoc:: gridfs
    """

    __motor_class_name__ = 'MongoMotorGridFS'

    def __init__(self, database, collection="fs"):
        """An instance of GridFS on top of a single Database.

        :Parameters:
          - `database`: a :class:`~mongomotor.MongoMotorDatabase`
          - `collection` (optional): A string, name of root collection to use,
            such as "fs" or "my_files"

        .. mongodoc:: gridfs

        """

        db_class = create_class_with_framework(
            MongoMotorAgnosticDatabase, database._framework, self.__module__)

        if not isinstance(database, db_class):
            raise TypeError("First argument to MongoMotorGridFS must be "
                            "MongoMotorDatabase, not %r" % database)

        self.io_loop = database.get_io_loop()
        self.collection = database[collection]
        self.delegate = self.__delegate_class__(
            database.delegate,
            collection)

    def wrap(self, obj):
        if obj.__class__ is grid_file.GridIn:
            grid_in_class = create_class_with_framework(
                MongoMotorAgnosticGridIn, self._framework, self.__module__)

            return grid_in_class(
                root_collection=self.collection,
                delegate=obj)

        elif obj.__class__ is grid_file.GridOut:
            grid_out_class = create_class_with_framework(
                MongoMotorAgnosticGridOut, self._framework, self.__module__)

            return grid_out_class(
                root_collection=self.collection,
                delegate=obj)

        elif obj.__class__ is gridfs.GridOutCursor:
            grid_out_class = create_class_with_framework(
                AgnosticGridOutCursor, self._framework, self.__module__)

            return grid_out_class(
                cursor=obj,
                collection=self.collection)



# -*- coding: utf-8 -*-

# Copyright 2016-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of mongomotor.

# mongomotor is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mongomotor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mongomotor. If not, see <http://www.gnu.org/licenses/>.

from uuid import uuid4
from mongoengine import fields
from mongoengine.base.datastructures import (
    BaseDict, BaseList, EmbeddedDocumentList)
from mongoengine.connection import get_db

from motor.metaprogramming import create_class_with_framework
from mongoengine.document import EmbeddedDocument
from mongomotor import gridfs
from mongomotor.metaprogramming import (asynchronize, Async,
                                        AsyncGenericMetaclass)

from mongoengine.fields import *  # noqa f403 for the sake of the api


class BaseAsyncReferenceField:
    """Base class to asynchronize reference fields."""

    def __get__(self, instance, owner):
        # When we are getting the field from a class not from an
        # instance we don't need a Future
        if instance is None:
            return self

        meth = super().__get__
        if self._auto_dereference:
            if isinstance(instance, EmbeddedDocument):
                # It's used when there's a reference in a EmbeddedDocument.
                # We use it to get the async framework to be used.
                instance._get_db = self.document_type._get_db
            meth = asynchronize(meth)

        return meth(instance, owner)


class ReferenceField(BaseAsyncReferenceField, fields.ReferenceField):
    """A reference to a document that will be automatically dereferenced on
    access (lazily).

    Use the `reverse_delete_rule` to handle what should happen if the document
    the field is referencing is deleted.  EmbeddedDocuments, DictFields and
    MapFields does not support reverse_delete_rule and an
    `InvalidDocumentError` will be raised if trying to set on one of these
    Document / Field types.

    The options are:

      * DO_NOTHING (0)  - don't do anything (default).
      * NULLIFY    (1)  - Updates the reference to null.
      * CASCADE    (2)  - Deletes the documents associated with the reference.
      * DENY       (3)  - Prevent the deletion of the reference object.
      * PULL       (4)  - Pull the reference from a
        :class:`~mongomotor.fields.ListField` of references

    Alternative syntax for registering delete rules (useful when implementing
    bi-directional delete rules)

    .. code-block:: python

        class Bar(Document):
            content = StringField()
            foo = ReferenceField('Foo')

        Foo.register_delete_rule(Bar, 'foo', NULLIFY)

    """

    pass


class GenericReferenceField(BaseAsyncReferenceField, fields.
                            GenericReferenceField):
    pass


class ComplexBaseField(fields.ComplexBaseField):

    def __get__(self, instance, owner):

        if instance is None:
            return self

        super_meth = super().__get__
        if isinstance(self.field, ReferenceField) and self._auto_dereference:
            r = asynchronize(super_meth)(instance, owner)
        else:
            r = super_meth(instance, owner)

        return r


class ListField(ComplexBaseField, fields.ListField):
    pass


class DictField(ComplexBaseField, fields.DictField):
    pass


class GridFSProxy(fields.GridFSProxy, metaclass=AsyncGenericMetaclass):

    delete = Async()
    new_file = Async()
    put = Async()
    read = Async()
    replace = Async()
    close = Async()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.grid_in = None
        self.grid_out = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc, exc_type, exc_tb):
        await self.close()

    @property
    def fs(self):
        if not self._fs:
            db = get_db(self.db_alias)
            grid_class = create_class_with_framework(
                gridfs.MongoMotorAgnosticGridFS, db._framework,
                'mongomotor.gridfs')

            self._fs = grid_class(db, self.collection_name)

        return self._fs

    def new_file(self, **metadata):
        """Opens a new stream for writing to gridfs.

        :param metadata: File's metadata.
        """
        file_name = uuid4().hex
        grid_in = self.fs.open_upload_stream(file_name, metadata=metadata)
        self.grid_id = grid_in._id
        self.grid_in = grid_in
        self._mark_as_changed()
        return self

    async def close(self):
        if self.grid_in:
            await self.grid_in.close()
            self.grid_in = None

    async def write(self, data):
        """Writes ``data`` to gridfs.

        :param data: String or bytes to write to gridfs."""

        if self.grid_id:
            if not self.grid_in:
                raise GridFSError(  # noqa f405
                    'This document already has a file. Either '
                    'delete it or call replace to overwrite it')

        elif not self.grid_in:
            raise GridFSError('You must create a new file first. Call '
                              '``new_file`` or use the async context manager')

        return self.grid_in.write(data)

    async def put(self, data, **metadata):
        """Writes ``data`` to gridfs.

        :param data: byte-string to write.
        :param metatada: File's metadata.
        """

        async with self.new_file(**metadata):
            await self.write(data)
        self._mark_as_changed()

    async def read(self):
        if not self.grid_id:
            return None

        self.grid_out = await self.fs.open_download_stream(self.grid_id)
        r = await self.grid_out.read()
        return r

    async def delete(self):
        # Delete file from GridFS, FileField still remains
        await self.fs.delete(self.grid_id)
        self.grid_in = None
        self.grid_id = None
        self.grid_out = None
        self._mark_as_changed()

    async def replace(self, data, **metadata):
        """Replaces the contents of the file with ``data``.

        :param data: A byte-string to write to gridfs.
        :param metatada: File metadata.
        """
        await self.delete()
        await self.put(data, **metadata)


class FileField(fields.FileField):

    proxy_class = GridFSProxy
